/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

//#define DEBUG_ENABLED
#include <audioscale.h>

/* elementfactory information */
static GstElementDetails audioscale_details = {
  "Audio scaler",
  "Filter/Audio/Scaler",
  "Resizes audio",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

/* Audioscale signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_FREQUENCY,
  /* FILL ME */
};



static void	gst_audioscale_class_init	(AudioscaleClass *klass);
static void	gst_audioscale_init		(Audioscale *audioscale);

static void	gst_audioscale_chain		(GstPad *pad, GstBuffer *buf);

static void	gst_audioscale_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_audioscale_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static GstElementClass *parent_class = NULL;
//static guint gst_audioscale_signals[LAST_SIGNAL] = { 0 };

GtkType
audioscale_get_type (void)
{
  static GtkType audioscale_type = 0;

  if (!audioscale_type) {
    static const GtkTypeInfo audioscale_info = {
      "Audioscale",
      sizeof(Audioscale),
      sizeof(AudioscaleClass),
      (GtkClassInitFunc)gst_audioscale_class_init,
      (GtkObjectInitFunc)gst_audioscale_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    audioscale_type = gtk_type_unique(GST_TYPE_ELEMENT,&audioscale_info);
  }
  return audioscale_type;
}

static void
gst_audioscale_class_init (AudioscaleClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtk_object_add_arg_type("Audioscale::frequency", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_FREQUENCY);

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_audioscale_set_arg;
  gtkobject_class->get_arg = gst_audioscale_get_arg;

}

static void
gst_audioscale_init (Audioscale *audioscale)
{
  audioscale->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(audioscale),audioscale->sinkpad);
  gst_pad_set_chain_function(audioscale->sinkpad,gst_audioscale_chain);
  audioscale->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(audioscale),audioscale->srcpad);

  audioscale->targetfrequency = -1;
  audioscale->buffer = NULL;
  audioscale->buffersize = 0;
}


static void
gst_audioscale_chain (GstPad *pad, GstBuffer *buf)
{
  Audioscale *audioscale;
  GstMeta *meta;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;
  int i;
  int pos;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  audioscale = GST_AUDIOSCALE (gst_pad_get_parent (pad));
  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"gst_audioscale_chain: got buffer of %ld bytes in '%s'\n",size,
		              gst_element_get_name (audioscale));

  /* first deal with audio metadata */
  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (audioscale->format != ((MetaAudioRaw *)meta)->format ||
        audioscale->channels  != ((MetaAudioRaw *)meta)->channels ||
        audioscale->frequency != ((MetaAudioRaw *)meta)->frequency) {

      audioscale->format = ((MetaAudioRaw *)meta)->format;
      audioscale->channels  = ((MetaAudioRaw *)meta)->channels;
      audioscale->frequency = ((MetaAudioRaw *)meta)->frequency;

    }
  }
  if (audioscale->buffer == NULL) {
    audioscale->buffer = g_malloc(size);
    memcpy(audioscale->buffer, data, size);
    audioscale->buffersize = size;
  }
  else {
    audioscale->buffer = g_realloc(audioscale->buffer, audioscale->buffersize + size);
    memcpy(audioscale->buffer+audioscale->buffersize, data, size);
    audioscale->buffersize += size;
  }

  while (audioscale->buffersize >= audioscale->frequency) {
    outbuf = gst_buffer_new();
    GST_BUFFER_DATA(outbuf) = g_malloc(audioscale->targetfrequency);
    GST_BUFFER_SIZE(outbuf) = audioscale->targetfrequency;
    //memcpy(GST_BUFFER_DATA(outbuf), audioscale->buffer, audioscale->targetfrequency);

    for (i=0; i< audioscale->targetfrequency/4; i++) {
      pos = (i * audioscale->frequency)/audioscale->targetfrequency;
      //printf("%d %d\n", i, pos);
      *(((gulong *)GST_BUFFER_DATA(outbuf))+i) = *(((gulong *)audioscale->buffer)+pos);
    }
    memcpy(audioscale->buffer, audioscale->buffer+audioscale->frequency, audioscale->buffersize-audioscale->frequency);
    audioscale->buffer = g_realloc(audioscale->buffer, audioscale->buffersize-audioscale->frequency);
    audioscale->buffersize -= audioscale->frequency;

    GST_DEBUG (0,"gst_audioscale_chain: pushing buffer of %d bytes in '%s'\n",GST_BUFFER_SIZE(outbuf),
		              gst_element_get_name (audioscale));

    gst_pad_push(audioscale->srcpad, outbuf);
  }
  gst_buffer_unref(buf);
}

static void
gst_audioscale_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  Audioscale *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_AUDIOSCALE(object));
  src = GST_AUDIOSCALE(object);

  switch(id) {
    case ARG_FREQUENCY:
      src->targetfrequency = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void
gst_audioscale_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  Audioscale *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_AUDIOSCALE(object));
  src = GST_AUDIOSCALE(object);

  switch(id) {
    case ARG_FREQUENCY:
      GTK_VALUE_INT(*arg) = src->targetfrequency;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}


GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("audioscale");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the audioscale element */
  factory = gst_elementfactory_new("audioscale",GST_TYPE_AUDIOSCALE,
                                   &audioscale_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
