/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_AVI_DECODER_H__
#define __GST_AVI_DECODER_H__


#include <config.h>
#include <gst/gst.h>
#include <gst/meta/audioraw.h>
#include <gst/meta/videoraw.h>
#include <libs/riff/gstriff.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_AVI_DECODER \
  (gst_avi_decoder_get_type())
#define GST_AVI_DECODER(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_AVI_DECODER,GstAviDecoder))
#define GST_AVI_DECODER_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_AVI_DECODER,GstAviDecoder))
#define GST_IS_AVI_DECODER(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_AVI_DECODER))
#define GST_IS_AVI_DECODER_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_AVI_DECODER))


#define GST_AVI_DECODER_UNKNOWN	  	0	/* initialized state */
#define GST_AVI_DECODER_REGULAR	  	1	/* regular parsing */
#define GST_AVI_DECODER_HDRL    	2
#define GST_AVI_DECODER_STRL    	3
#define GST_AVI_DECODER_MOVI    	4
#define GST_AVI_DECODER_AVIH    	5
#define GST_AVI_DECODER_STRH_VIDS	6
#define GST_AVI_DECODER_STRH_AUDS	7

#define GST_AVI_DECODER_MAX_AUDIO_PADS	8	
#define GST_AVI_DECODER_MAX_VIDEO_PADS	8	

typedef struct _GstAviDecoder GstAviDecoder;
typedef struct _GstAviDecoderClass GstAviDecoderClass;

struct _GstAviDecoder {
  GstElement element;

  /* pads */
  GstPad *sinkpad,*srcpad;

  /* AVI decoding state */
  gint state;

  /* RIFF decoding state */
  GstRiff *riff;
  GstRiff *index;

  gst_riff_index_entry *index_entries;
  gulong index_size;
  gulong index_offset;
  gulong resync_offset;

  guint64 next_time;
  guint64 time_interval;
  gulong tot_frames;
  gulong current_frame;

  guint32 flags;

  guint num_audio_pads;
  guint num_video_pads;
  GstPad       *audio_pad[GST_AVI_DECODER_MAX_AUDIO_PADS];
  MetaAudioRaw *audio_meta[GST_AVI_DECODER_MAX_AUDIO_PADS];
  gboolean 	audio_need_flush[GST_AVI_DECODER_MAX_AUDIO_PADS];

  GstPad       *video_pad[GST_AVI_DECODER_MAX_VIDEO_PADS];
  MetaVideoRaw *video_meta[GST_AVI_DECODER_MAX_VIDEO_PADS];
  gboolean 	video_need_flush[GST_AVI_DECODER_MAX_VIDEO_PADS];

  gpointer extra_data;
};

struct _GstAviDecoderClass {
  GstElementClass parent_class;
};

GtkType 		gst_avi_decoder_get_type		(void);
GstElement*		gst_avi_decoder_new			(gchar *name);

GstPad*			gst_avi_decoder_get_video_srcpad	(GstAviDecoder *avi_decoder, 
								 guint pad_nr, gst_riff_strf_vids *strf,
								 GstPadTemplate *temp);
GstPad*			gst_avi_decoder_get_audio_srcpad	(GstAviDecoder *avi_decoder, 
								 guint pad_nr, gst_riff_strf_auds *strf,
								 GstPadTemplate *temp);

void 			gst_avi_decoder_winloader_handle	(GstAviDecoder *avi_decoder, GstBuffer *buf);
void 			gst_avi_decoder_winloader_chain		(GstPad *pad, GstBuffer *buf);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_AVI_DECODER_H__ */
