/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



//#define DEBUG_ENABLED

#include "config.h"
#ifdef HAVE_LIBMMX
#include "mmx.h"
#endif

#include <gstavidecoder.h>

static GstPad *gst_avi_decoder_get_srcpad_MJPEG(GstAviDecoder *avi_decoder, guint pad_nr, GstPadTemplate *temp);
static GstPad *gst_avi_decoder_get_srcpad_winloader(GstAviDecoder *avi_decoder, guint pad_nr, gst_riff_strf_vids *strf, GstPadTemplate *temp);

GstPad *gst_avi_decoder_get_video_srcpad(GstAviDecoder *avi_decoder, guint pad_nr, gst_riff_strf_vids *strf, GstPadTemplate *temp) 
{
  GstPad *newpad;
  GstMeta *meta;

  switch (strf->compression) {
    case GST_RIFF_MJPG:
      return gst_avi_decoder_get_srcpad_MJPEG(avi_decoder, pad_nr, temp); 
      break;
    case GST_RIFF_I420:
      meta = gst_meta_new_size(sizeof(MetaVideoRaw));
      ((MetaVideoRaw *)meta)->format = GST_COLORSPACE_YUV420P;
      ((MetaVideoRaw *)meta)->width = strf->width;
      ((MetaVideoRaw *)meta)->height = strf->height;
      newpad = gst_pad_new_from_template(temp, "video_00");
      avi_decoder->video_pad[pad_nr] = newpad;
      avi_decoder->video_meta[pad_nr] = (MetaVideoRaw *)meta;
      return newpad;
    case GST_RIFF_YV12:
    default:
      newpad = gst_avi_decoder_get_srcpad_winloader(avi_decoder, pad_nr, strf, temp); 
      if (newpad) return newpad;
      printf("codec %08x not supported\n", strf->compression);
      break;
  }
  return NULL;
}

static GstPad *gst_avi_decoder_get_srcpad_MJPEG(GstAviDecoder *avi_decoder, guint pad_nr, GstPadTemplate *temp) 
{
  GstElementFactory *jpegfactory;
  GstElement *jpeg;
  GstPad *srcpad, *sinkpad, *newpad;

  jpegfactory = gst_elementfactory_find("jpegdec");
  g_return_val_if_fail(jpegfactory != NULL, NULL);

  jpeg = gst_elementfactory_create(jpegfactory, "jpegdec");
  gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(avi_decoder))), GST_STATE_PAUSED);
  gst_bin_add(GST_BIN(gst_object_get_parent(GST_OBJECT(avi_decoder))), jpeg);

  newpad = gst_pad_new("video", GST_PAD_SRC);
  gst_pad_set_parent(newpad, GST_OBJECT(avi_decoder));

  sinkpad = gst_element_get_pad(jpeg,"sink");
  srcpad = gst_element_get_pad(jpeg,"src");

  gst_pad_connect(newpad, sinkpad);
  gst_pad_set_name(srcpad, "video_00");
  gst_pad_set_chain_function (newpad, GST_RPAD_CHAINFUNC (sinkpad));

  printf("playcodecs: filling pad %p %d %p %p\n", avi_decoder, pad_nr, srcpad, &avi_decoder->video_pad[pad_nr]);
  avi_decoder->video_pad[pad_nr] = newpad;
  avi_decoder->video_meta[pad_nr] = NULL;

  gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(avi_decoder))), GST_STATE_PLAYING);
  return srcpad;
}

static GstPad *gst_avi_decoder_get_srcpad_winloader(GstAviDecoder *avi_decoder, guint pad_nr, gst_riff_strf_vids *strf, GstPadTemplate *temp) 
{
  GstElementFactory *windecfactory;
  GstElement *windec;
  GstPad *srcpad, *sinkpad, *newpad;

  windecfactory = gst_elementfactory_find("windec");
  g_return_val_if_fail(windecfactory != NULL, NULL);

  windec = gst_elementfactory_create(windecfactory, "windec");
  g_return_val_if_fail(windec != NULL, NULL);
  gtk_object_set(GTK_OBJECT(windec), "avi_header", strf, NULL);
  gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(avi_decoder))), GST_STATE_PAUSED);
  gst_bin_add(GST_BIN(gst_object_get_parent(GST_OBJECT(avi_decoder))), windec);

  newpad = gst_pad_new("video", GST_PAD_SRC);
  gst_pad_set_parent(newpad, GST_OBJECT(avi_decoder));

  sinkpad = gst_element_get_pad(windec,"sink");
  srcpad = gst_element_get_pad(windec,"src");

  gst_pad_connect(newpad, sinkpad);
  gst_pad_set_name(srcpad, "video_00");
  gst_pad_set_chain_function (newpad, GST_RPAD_CHAINFUNC (sinkpad));

  printf("playcodecs: filling pad %p %d %p %p\n", avi_decoder, pad_nr, srcpad, &avi_decoder->video_pad[pad_nr]);
  avi_decoder->video_pad[pad_nr] = newpad;
  avi_decoder->video_meta[pad_nr] = NULL;

  gst_element_set_state(GST_ELEMENT(gst_object_get_parent(GST_OBJECT(avi_decoder))), GST_STATE_PLAYING);
  return srcpad;
}

