/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "gstwindec.h"
#include "gstwinenc.h"

static GstPadFactory src_factory = {
  "src",
  GST_PAD_FACTORY_SRC,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS (
  "wincodec_src",
    "video/raw",
    "fourcc",    GST_PROPS_LIST (
                   GST_PROPS_FOURCC ('Y','V','1','2'),
                   GST_PROPS_FOURCC ('I','Y','U','V'),
                   GST_PROPS_FOURCC ('I','4','2','0')
                 ),
    "width",     GST_PROPS_INT_RANGE (16, 4096),
    "height",    GST_PROPS_INT_RANGE (16, 4096)
  ),
  NULL,
};

/* elementfactory information */
extern GstElementDetails gst_windec_details;
extern GstElementDetails gst_winenc_details;

GstPadTemplate *src_temp;

GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *dec, *enc;

  if (!gst_library_load("winloader")) {
    gst_info("playcodecs: could not load support library: 'winloader'\n");
    return NULL;
  }
  if (!gst_library_load("gstcolorspace")) {
    gst_info("playcodecs: could not load support library: 'gstcolorspace'\n");
    return NULL;
  }

  plugin = gst_plugin_new("wincodec");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the windec element */
  enc = gst_elementfactory_new("winenc",GST_TYPE_WINENC,
                                   &gst_winenc_details);
  g_return_val_if_fail(enc != NULL, NULL); 
  gst_plugin_add_factory(plugin,enc);

  /* create an elementfactory for the windec element */
  dec = gst_elementfactory_new("windec",GST_TYPE_WINDEC,
                                   &gst_windec_details);
  g_return_val_if_fail(dec != NULL, NULL); 

  src_temp = gst_padtemplate_new (&src_factory);
  gst_elementfactory_add_padtemplate (dec, src_temp);
  
  gst_plugin_add_factory(plugin,dec);

  return plugin;
}
