/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

//#define DEBUG_ENABLED
#define WIN32

#include "gstwinenc.h"
#include <libs/riff/gstriff.h>

#include "config.h"
#ifdef HAVE_LIBMMX
#include "mmx.h"
#endif

#include <wine/driver.h>
#include <wine/vfw.h>
#include <libs/winloader/registry.h>



/* elementfactory information */
GstElementDetails gst_winenc_details = {
  "Windows codec image encoder",
  "Filter/Encoder/Image",
  NULL,
  VERSION,
  "Wim Taymans <wim.taymans@tvd.be>",
  "(C) 2000",
};

#define HKEY_CLASSES_ROOT       (0x80000000)
#define HKEY_CURRENT_USER       (0x80000001)
#define HKEY_LOCAL_MACHINE      (0x80000002)
#define HKEY_USERS              (0x80000003)
#define HKEY_PERFORMANCE_DATA   (0x80000004)
#define HKEY_CURRENT_CONFIG     (0x80000005)
#define HKEY_DYN_DATA           (0x80000006)
#define REG_DWORD               4       /* DWORD in little endian format */


/* WinEnc signals and args */
enum {
  /* FILL ME */
  SIGNAL_FRAME_ENCODED,
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_AVI_HEADER,
  ARG_BITRATE,
  ARG_QUALITY,
  ARG_COMPRESSION,
  ARG_LAST_FRAME_SIZE,
  /* FILL ME */
};

typedef struct _GstWinLoaderData GstWinLoaderData;

struct _GstWinLoaderData {
  COMPVARS m_cv;
  BITMAPINFOHEADER m_obh;
  BITMAPINFOHEADER m_bh;
  gint m_iFrameNum;
  BITMAPINFOHEADER m_prevbh;
  guchar *m_prev;
};

static void gst_winenc_class_init(GstWinEnc *klass);
static void gst_winenc_init(GstWinEnc *winenc);

GstElement *gst_winenc_new(gchar *name);
static void gst_winenc_chain(GstPad *pad,GstBuffer *buf);
static gboolean gst_winenc_init_encoder(GstWinEnc *winenc);

static void gst_winenc_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_winenc_get_arg(GtkObject *object,GtkArg *arg,guint id);

static GstElementClass *parent_class = NULL;
static guint gst_winenc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_winenc_get_type(void) {
  static GtkType winenc_type = 0;

  if (!winenc_type) {
    static const GtkTypeInfo winenc_info = {
      "GstWinEnc",
      sizeof(GstWinEnc),
      sizeof(GstWinEnc),
      (GtkClassInitFunc)gst_winenc_class_init,
      (GtkObjectInitFunc)gst_winenc_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    winenc_type = gtk_type_unique(GST_TYPE_ELEMENT,&winenc_info);
  }
  return winenc_type;
}

static void
gst_winenc_class_init(GstWinEnc *klass) {
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstWinEnc::avi_header", GTK_TYPE_POINTER,
                               GTK_ARG_WRITABLE, ARG_AVI_HEADER);
  gtk_object_add_arg_type("GstWinEnc::bitrate", GTK_TYPE_INT,
                               GTK_ARG_WRITABLE, ARG_BITRATE);
  gtk_object_add_arg_type("GstWinEnc::quality", GTK_TYPE_INT,
                               GTK_ARG_WRITABLE, ARG_QUALITY);
  gtk_object_add_arg_type("GstWinEnc::compression", GTK_TYPE_LONG,
                               GTK_ARG_WRITABLE, ARG_COMPRESSION);
  gtk_object_add_arg_type("GstWinEnc::last_frame_size", GTK_TYPE_LONG,
                               GTK_ARG_READABLE, ARG_LAST_FRAME_SIZE);

  gtkobject_class->set_arg = gst_winenc_set_arg;
  gtkobject_class->get_arg = gst_winenc_get_arg;

  gst_winenc_signals[SIGNAL_FRAME_ENCODED] =
      gtk_signal_new("frame_encoded",GTK_RUN_LAST,gtkobject_class->type,
      GTK_SIGNAL_OFFSET(GstWinEncClass,frame_encoded),
      gtk_marshal_NONE__INT,GTK_TYPE_NONE,1,
      GTK_TYPE_INT);

  gtk_object_class_add_signals(gtkobject_class,gst_winenc_signals,
                                    LAST_SIGNAL);
}

static void gst_winenc_init(GstWinEnc *winenc) {
  /* create the sink and src pads */
  winenc->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(winenc),winenc->sinkpad);
  gst_pad_set_chain_function(winenc->sinkpad,gst_winenc_chain);
  winenc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(winenc),winenc->srcpad);

  // reset the initial video state
  winenc->format = GST_COLORSPACE_YUV420P;
  winenc->width = -1;
  winenc->height = -1;
  winenc->bitrate = 800;
  winenc->quality = 10000;
  winenc->compression = GST_RIFF_DIV3;

  winenc->last_frame_size = 0;

  winenc->inited = FALSE;

  winenc->extra_data = g_malloc0(sizeof(GstWinLoaderData));
}

GstElement*
gst_winenc_new (gchar *name)
{
  GstElement *winenc = GST_ELEMENT (gtk_type_new (GST_TYPE_WINENC));

  gst_element_set_name (winenc, name);

  return winenc;
}

static void gst_winenc_chain(GstPad *pad,GstBuffer *buf) {
  GstWinEnc *winenc;
  guchar *data, *outdata, *cnvbuf;
  gulong size;
  GstBuffer *outbuf;
  GstMeta *meta;
  gint result;
  glong lpckid, is_keyframe;
  GstWinLoaderData * extra_data;
  gint i, j;
  guint32 *ptr, *outptr;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  //g_return_if_fail(GST_IS_BUFFER(buf));

  winenc = GST_WINENC (gst_pad_get_parent (pad));

  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"gst_winenc_chain: got buffer of %ld bytes in '%s'\n",size,
          gst_element_get_name (GST_ELEMENT (winenc)));

  /* first deal with video metadata */
  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (winenc->format != ((MetaVideoRaw *)meta)->format ||
        winenc->width  != ((MetaVideoRaw *)meta)->width ||
        winenc->height != ((MetaVideoRaw *)meta)->height) {

      winenc->format = ((MetaVideoRaw *)meta)->format;
      winenc->width  = ((MetaVideoRaw *)meta)->width;
      winenc->height = ((MetaVideoRaw *)meta)->height;
    }
  }

  if (!winenc->inited) {
    gst_winenc_init_encoder(winenc);
    winenc->inited = TRUE;
  }

  extra_data = (GstWinLoaderData *)winenc->extra_data;

  cnvbuf = g_malloc(winenc->width * winenc->height *3);

  GST_DEBUG (0,"gst_winenc_chain: convert\n");
  if (winenc->converter) gst_colorspace_convert(winenc->converter, data, cnvbuf);
  GST_DEBUG (0,"gst_winenc_chain: convert done\n");

  ptr = (guint32 *) cnvbuf;
  outptr = (guint32 *) (cnvbuf + winenc->width * winenc->height *3);
  outptr -= (extra_data->m_obh.biWidth * 3)/4;

  for (i=extra_data->m_obh.biHeight/2; i; i--) {
    for (j=(extra_data->m_obh.biWidth*3)/8; j; j--) {
#ifdef HAVE_LIBMMX
      movq_m2r(*ptr, mm0);
      movq_m2r(*outptr, mm1);
      movq_r2m(mm1, *ptr);
      movq_r2m(mm0, *outptr);
      ptr+=2;
      outptr+=2;
#else
      guint32 temp = *ptr;
      *ptr++ = *outptr;
      *outptr++ = temp;
      temp = *ptr;
      *ptr++ = *outptr;
      *outptr++ = temp;
#endif
    }
    outptr -= (extra_data->m_obh.biWidth * 3)/2;
  }
#ifdef HAVE_LIBMMX
  emms();
#endif

  outbuf = gst_buffer_new();
  outdata = GST_BUFFER_DATA(outbuf) = g_malloc(1280000);

  result=ICCompress(extra_data->m_cv.hic, ICCOMPRESS_KEYFRAME,  &extra_data->m_obh, outdata, &extra_data->m_bh, cnvbuf,
	             (long*)&lpckid, (long*)&is_keyframe, extra_data->m_iFrameNum, 0, winenc->quality, &extra_data->m_prevbh, extra_data->m_prev);
  g_free(cnvbuf);

  if(result==0)
  {
    if(extra_data->m_prev == NULL) extra_data->m_prev = g_malloc(extra_data->m_bh.biSizeImage);

    memcpy(extra_data->m_prev, data, extra_data->m_bh.biSizeImage);
    memcpy(&extra_data->m_prevbh, &extra_data->m_bh, sizeof(extra_data->m_bh));
  }    
	
  winenc->last_frame_size = GST_BUFFER_SIZE(outbuf) = extra_data->m_obh.biSizeImage;
  GST_DEBUG (0,"gst_winenc_chain: compressed %ld, %ld %d %d %d\n",
             lpckid, is_keyframe, extra_data->m_iFrameNum, result, GST_BUFFER_SIZE(outbuf));
  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);

  gtk_signal_emit(GTK_OBJECT(winenc),gst_winenc_signals[SIGNAL_FRAME_ENCODED],
                                            extra_data->m_iFrameNum);

  extra_data->m_iFrameNum++;

  gst_pad_push(winenc->srcpad, outbuf);

  gst_buffer_unref(buf);
}

static gboolean gst_winenc_init_encoder(GstWinEnc *winenc) 
{
  HRESULT h;
  GstWinLoaderData *data;
  gint result, status, newkey;

  data = (GstWinLoaderData *)winenc->extra_data;

  data->m_iFrameNum = 0;

  result=RegCreateKeyExA(HKEY_CURRENT_USER, "Software\\LinuxLoader\\Divx", 0, 0, 0, 0, 0,
          &newkey, &status);
  if(result!=0) {
     printf("gstwinenc:: RegCreateKeyExA failed\n");
     return FALSE;
  }
  status = winenc->bitrate;
  result=RegSetValueExA(newkey, "BitRate", 0, REG_DWORD, &status, 4);
  if(result!=0) {
     printf("gstwinenc:: RegSetValueExA failed\n");
     return FALSE;
  }

  winenc->converter = gst_colorspace_converter_new(winenc->width, winenc->height, winenc->format, GST_COLORSPACE_BGR24, NULL);

  data->m_bh.biSize = sizeof(data->m_bh);
  data->m_bh.biWidth = winenc->width;
  data->m_bh.biHeight = winenc->height;
  data->m_bh.biPlanes = 1;
  data->m_bh.biBitCount = 24;
  data->m_bh.biCompression = GST_RIFF_rgb;
  data->m_bh.biSizeImage = data->m_bh.biWidth * data->m_bh.biHeight * data->m_bh.biBitCount/8;
  data->m_bh.biXPelsPerMeter = 0;
  data->m_bh.biYPelsPerMeter = 0;
  data->m_bh.biClrUsed = 0;
  data->m_bh.biClrImportant = 0;
  
  memset(&data->m_obh, 0, sizeof(data->m_obh));
  data->m_obh.biSize=sizeof(data->m_obh);

  memset(&data->m_cv, 0, sizeof(data->m_cv));
  data->m_cv.cbSize=sizeof(data->m_cv);
  gst_info("gstwinenc: trying to open library %p\n", data);
  data->m_cv.hic=ICOpen(
           GST_RIFF_FCC_vidc,
           winenc->compression,
           ICMODE_COMPRESS);

  data->m_cv.fccType=GST_RIFF_FCC_vidc,
  data->m_cv.fccHandler=winenc->compression;

  if(data->m_cv.hic==0)
  {
     printf("gstwinenc:: ICOpen failed\n");
     return FALSE;
  }

  gst_info("gstwinenc: get format \n");
  
  h=ICCompressGetFormat(data->m_cv.hic, &data->m_bh, &data->m_obh);
  if(h!=0)
  {
    printf("gstwinenc:: Can't handle this format %ld\n", h);
    ICClose(data->m_cv.hic);
    return FALSE;
  }

  h=ICCompressBegin(data->m_cv.hic, &data->m_bh, &data->m_obh);
  if(h!=0)
  {
    printf("gstwinenc:: ICCompressBegin failed\n");
    return FALSE;
  }

  return TRUE;
}

static void gst_winenc_set_arg(GtkObject *object,GtkArg *arg,guint id)
{
  GstWinEnc *winenc;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINENC(object));
  winenc = GST_WINENC(object);

  data = (GstWinLoaderData *)winenc->extra_data;

  switch(id) {
    case ARG_AVI_HEADER:
      memcpy(&data->m_bh, GTK_VALUE_POINTER(*arg), sizeof(BITMAPINFOHEADER));
      break;
    case ARG_BITRATE:
      winenc->bitrate = GTK_VALUE_INT(*arg);
      break;
    case ARG_QUALITY:
      winenc->quality = GTK_VALUE_INT(*arg);
      break;
    case ARG_COMPRESSION:
      winenc->compression = GTK_VALUE_LONG(*arg);
      break;
    default:
      break;
  }
}

static void gst_winenc_get_arg(GtkObject *object,GtkArg *arg,guint id)
{
  GstWinEnc *winenc;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINENC(object));
  winenc = GST_WINENC(object);

  data = (GstWinLoaderData *)winenc->extra_data;

  switch(id) {
    case ARG_LAST_FRAME_SIZE:
      GTK_VALUE_LONG(*arg) = winenc->last_frame_size;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

