/* GStreamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <math.h>
#include <linux/soundcard.h>

//#define DEBUG_ENABLED
#include "gstladspa.h"
#include "ladspa.h"
#include "search.h"
#include "utils.h"


/* FIXME
static GstCapsFactory audio_raw_caps = {
  "audio/raw",
  "format",   GST_PROPS_INT (AFMT_S16_LE),
  "float",    GST_PROPS_LIST (
                GST_PROPS_BOOL (TRUE),
                GST_PROPS_BOOL (FALSE)
              ),
  "depth",    GST_PROPS_INT_RANGE (8, 32),
  "channels", GST_PROPS_INT (1),
  NULL
};

static GstCapsFactory audio_float_caps = {
  "audio/float",
  "channels", GST_PROPS_INT (1),
  NULL
};
*/

enum {
  ARG_0,
  ARG_LOOP_BASED,
  ARG_LAST,
};

static void			gst_ladspa_class_init		(GstLADSPAClass *klass);
static void			gst_ladspa_init			(GstLADSPA *ladspa);

static void			gst_ladspa_set_arg		(GtkObject *object,GtkArg *arg,guint id);
static void			gst_ladspa_get_arg		(GtkObject *object,GtkArg *arg,guint id);

static GstElementStateReturn	gst_ladspa_change_state		(GstElement *element);
static void			gst_ladspa_chain		(GstPad *pad,GstBuffer *buf);
static void			gst_ladspa_loop			(GstElement *element);


static GstElementClass *parent_class = NULL;
//static guint gst_ladspa_signals[LAST_SIGNAL] = { 0 };

static GstPlugin *ladspa_plugin;
static GHashTable *ladspa_descriptors;

static void
gst_ladspa_class_init (GstLADSPAClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;
  LADSPA_Descriptor *desc;
  gint i,j;
  gint hintdesc;
  gint argtype,argperms;
  gchar *argname;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtkobject_class->set_arg = gst_ladspa_set_arg;
  gtkobject_class->get_arg = gst_ladspa_get_arg;

  gstelement_class->change_state = gst_ladspa_change_state;

  gtk_object_add_arg_type (g_strdup_printf("%s::loop_based",gtk_type_name(gtkobject_class->type)),
                           GTK_TYPE_BOOL, GTK_ARG_READWRITE, ARG_LOOP_BASED);

  // look up and store the ladspa descriptor
  klass->descriptor = g_hash_table_lookup(ladspa_descriptors,
    GINT_TO_POINTER(gtkobject_class->type));
  desc = klass->descriptor;

  // allocate the controls struct
  klass->numcontrols = desc->PortCount;
  klass->controls = g_new0(ladspa_control,klass->numcontrols);

  // walk through the ports and add all the arguments
  for (i=0;i<desc->PortCount;i++) {
    if (LADSPA_IS_PORT_CONTROL(desc->PortDescriptors[i])) {
      // short name for hint descriptor
      hintdesc = desc->PortRangeHints[i].HintDescriptor;

      // grab the bounds
      if (LADSPA_IS_HINT_BOUNDED_BELOW(hintdesc)) {
        klass->controls[i].lower = TRUE;
        klass->controls[i].lowerbound =
          desc->PortRangeHints[i].LowerBound;
      }
      if (LADSPA_IS_HINT_BOUNDED_ABOVE(hintdesc)) {
        klass->controls[i].upper = TRUE;
        klass->controls[i].upperbound =
          desc->PortRangeHints[i].UpperBound;
        if (LADSPA_IS_HINT_SAMPLE_RATE(hintdesc))
          klass->controls[i].upperrate = TRUE;
      }

      // get the various bits
      if (LADSPA_IS_HINT_TOGGLED(hintdesc))
        klass->controls[i].toggled = TRUE;
      if (LADSPA_IS_HINT_LOGARITHMIC(hintdesc))
        klass->controls[i].logarithmic = TRUE;
      if (LADSPA_IS_HINT_INTEGER(hintdesc))
        klass->controls[i].integer = TRUE;

      // figure out the argument details
      if (klass->controls[i].toggled) argtype = GTK_TYPE_BOOL;
      else if (klass->controls[i].integer) argtype = GTK_TYPE_INT;
      else argtype = GTK_TYPE_FLOAT;

      if (LADSPA_IS_PORT_INPUT(desc->PortDescriptors[i])) {
        argperms = GTK_ARG_READWRITE;
        klass->controls[i].writable = TRUE;
      } else {
        argperms = GTK_ARG_READABLE;
        klass->controls[i].writable = FALSE;
      }

      // figure out what to call this thing
      argname = g_strdup_printf("%s::%s",
        gtk_type_name(gtkobject_class->type),
        desc->PortNames[i]);
      for (j=0;j<strlen(argname);j++)
        if (argname[j] == ' ') argname[j] = '_';

      // now we can go register the argument
      gtk_object_add_arg_type(argname,argtype,argperms,i+ARG_LAST);
      fprintf(stderr,"added arg %s\n",argname);
    }
  }
}

static void
gst_ladspa_init (GstLADSPA *ladspa)
{
  GstLADSPAClass *oclass = (GstLADSPAClass*)(GTK_OBJECT(ladspa)->klass);
  LADSPA_Descriptor *desc;
  gint i,sinkcount,srccount;

  desc = oclass->descriptor;
  ladspa->descriptor = oclass->descriptor;

  // walk through the ports and count the pads needed
  ladspa->numsinkpads = 0;
  ladspa->numsrcpads = 0;
  for (i=0;i<desc->PortCount;i++) {
    if (LADSPA_IS_PORT_AUDIO(desc->PortDescriptors[i])) {
      if (LADSPA_IS_PORT_INPUT(desc->PortDescriptors[i]))
        ladspa->numsinkpads++;
      else if (LADSPA_IS_PORT_OUTPUT(desc->PortDescriptors[i]))
        ladspa->numsrcpads++;
    }
  }

  // allocate the various arrays
  ladspa->srcpads = g_new0(GstPad*,ladspa->numsrcpads);
  ladspa->sinkpads = g_new0(GstPad*,ladspa->numsinkpads);
  ladspa->inputbuffers = g_new0(GstBuffer*,ladspa->numsinkpads);

  // walk through the ports and add all the pads
  sinkcount = 0;
  srccount = 0;
  for (i=0;i<desc->PortCount;i++) {
    if (LADSPA_IS_PORT_AUDIO(desc->PortDescriptors[i])) {
      if (LADSPA_IS_PORT_INPUT(desc->PortDescriptors[i])) {
        ladspa->sinkpads[sinkcount] = gst_pad_new(
          (gchar *)desc->PortNames[i],GST_PAD_SINK);
        gst_pad_set_chain_function(ladspa->sinkpads[sinkcount],gst_ladspa_chain);
        gst_element_add_pad(GST_ELEMENT(ladspa),ladspa->sinkpads[sinkcount]);
        sinkcount++;
      } else {
        ladspa->srcpads[srccount] = gst_pad_new(
          (gchar *)desc->PortNames[i],GST_PAD_SRC);
        gst_element_add_pad(GST_ELEMENT(ladspa),ladspa->srcpads[srccount]);
        srccount++;
      }
    }
  }

  ladspa->samplerate = 44100;
}

static void
gst_ladspa_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstLADSPA *ladspa = (GstLADSPA*)object;
  gint cid = id - 1;
  GstLADSPAClass *oclass = (GstLADSPAClass*)(object->klass);
  ladspa_control *control;
  gfloat val;

  if (id == ARG_LOOP_BASED) {
    // we can only do this in NULL state
    g_return_if_fail (GST_STATE(object) != GST_STATE_NULL);
    ladspa->loopbased = GTK_VALUE_BOOL (*arg);
    if (ladspa->loopbased) {
      gst_element_set_loop_function (GST_ELEMENT (ladspa), gst_ladspa_loop);
    } else {
      gst_element_set_loop_function (GST_ELEMENT (ladspa), NULL);
    }
  }

  // verify it exists and is a control (not a port)
  if (cid >= oclass->numcontrols) return;
  control = &(oclass->controls[cid]);
  if (control->name == NULL) return;

  // check to see if it's writable
  if (control->writable == FALSE) return;

  // now see what type it is
  if (control->toggled) {
    if (GTK_VALUE_BOOL(*arg))
      ladspa->controls[cid] = 1.0;
    else
      ladspa->controls[cid] = 0.0;
  } else if (control->integer) {
    val = (gfloat)GTK_VALUE_INT(*arg);
    if (control->lower && (val < control->lowerbound)) return;
    if (control->upper && (val > control->upperbound)) return;
    ladspa->controls[cid] = val;
  } else {
    val = GTK_VALUE_FLOAT(*arg);
    if (control->lower && (val < control->lowerbound)) return;
    if (control->upper && (val > control->upperbound)) return;
    ladspa->controls[cid] = val;
  }

fprintf(stderr,"set arg %s to %f\n",control->name,ladspa->controls[cid]);
}

static void
gst_ladspa_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstLADSPA *ladspa = (GstLADSPA*)object;
  gint cid = id - 1;
  GstLADSPAClass *oclass = (GstLADSPAClass*)(object->klass);
  ladspa_control *control;

  // verify it exists and is a control (not a port)
  if (cid >= oclass->numcontrols) return;
  control = &(oclass->controls[cid]);
  if (control->name == NULL) return;

  // now see what type it is
  if (control->toggled) {
    if (ladspa->controls[cid] == 1.0)
      GTK_VALUE_BOOL(*arg) = TRUE;
    else
      GTK_VALUE_BOOL(*arg) = FALSE;
  } else if (control->integer) {
    GTK_VALUE_INT(*arg) = (gint)ladspa->controls[cid];
  } else {
    GTK_VALUE_FLOAT(*arg) = ladspa->controls[cid];
  }

fprintf(stderr,"got arg %s as %f\n",control->name,ladspa->controls[cid]);
}

static gboolean
gst_ladspa_instantiate (GstLADSPA *ladspa)
{
  LADSPA_Descriptor *desc;
  int i;

  desc = ladspa->descriptor;

  // instantiate the plugin
  ladspa->handle = desc->instantiate(desc,ladspa->samplerate);
  g_return_val_if_fail (ladspa->handle != NULL, FALSE);

  // walk through the ports and add all the arguments
  for (i=0;i<desc->PortCount;i++) {
    if (LADSPA_IS_PORT_CONTROL(desc->PortDescriptors[i])) {
      // connect the argument to the plugin
      desc->connect_port(ladspa->handle,i,&(ladspa->controls[i]));
    }
  }

  // activate the plugin
  desc->activate(ladspa->handle);

  return TRUE;
}

static GstElementStateReturn
gst_ladspa_change_state (GstElement *element)
{
  GstLADSPA *ladspa = (GstLADSPA*)element;

  switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_NULL_TO_READY:
      if (!gst_ladspa_instantiate(ladspa))
        return GST_STATE_FAILURE;
      break;
    case GST_STATE_READY_TO_PLAYING:
      memset(ladspa->inputbuffers,0,sizeof(GstBuffer*) * ladspa->numsinkpads);
      ladspa->remainingpads = ladspa->numsinkpads;
      break;
    default:
      break;
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

static void
gst_ladspa_chain (GstPad *pad,GstBuffer *buf)
{
  int i;

  GstLADSPA *ladspa = (GstLADSPA*)GST_OBJECT_PARENT(pad);

  // find the pad number
  i=0;
  while (pad != ladspa->sinkpads[i]) i++;

  // store the buffer
  ladspa->inputbuffers[i] = buf;

  // decrement remaining count
  ladspa->remainingpads--;

  // now, do something
  if (ladspa->remainingpads == 0) {
    printf("doing something\n");

    memset(ladspa->inputbuffers,0,sizeof(GstBuffer*) * ladspa->numsinkpads);
    ladspa->remainingpads = ladspa->numsinkpads;
  }
}

static void
gst_ladspa_loop	(GstElement *element)
{
  GstLADSPA *ladspa = (GstLADSPA*)element;

  g_return_if_fail (ladspa != NULL);

  do {
    fprintf(stderr,"in ladspa_loop\n");
  } while (!GST_ELEMENT_IS_COTHREAD_STOPPING (element));
}


static void
ladspa_describe_plugin(const char *pcFullFilename,
                       void *pvPluginHandle,
                       LADSPA_Descriptor_Function pfDescriptorFunction)
{
  const LADSPA_Descriptor *psDescriptor;
  int i;
  GstElementDetails *details;
  GtkTypeInfo typeinfo = {
    NULL,
    sizeof(GstLADSPA),
    sizeof(GstLADSPAClass),
    (GtkClassInitFunc)gst_ladspa_class_init,
    (GtkObjectInitFunc)gst_ladspa_init,
    (GtkArgSetFunc)NULL,
    (GtkArgGetFunc)NULL,
    (GtkClassInitFunc)NULL,
  };
  GtkType type;
  GstElementFactory *factory;

  // walk through all the plugins in this pluginlibrary
  i = 0;
  while ((psDescriptor = pfDescriptorFunction(i++))) {
    // construct the type
    typeinfo.type_name =
      g_strdup_printf("LADSPA_%s",psDescriptor->Label);
    // if it's already registered, drop it
    if (gtk_type_from_name(typeinfo.type_name)) {
      g_free(typeinfo.type_name);
      continue;
    }
    // create the gtk type now
    type = gtk_type_unique(GST_TYPE_ELEMENT,&typeinfo);

    // construct the element details struct
    details = g_new0(GstElementDetails,1);
    details->longname = g_strdup(psDescriptor->Name);
    details->klass = "Filter/LADSPA";
    details->description = details->longname;
    details->version = g_strdup_printf("%ld",psDescriptor->UniqueID);
    details->author = g_strdup(psDescriptor->Maker);
    details->copyright = g_strdup(psDescriptor->Copyright);

    // register the plugin with gstreamer
    factory = gst_elementfactory_new(typeinfo.type_name,type,details);
    g_return_if_fail(factory != NULL);
    gst_plugin_add_factory(ladspa_plugin,factory);

    // add this plugin to the hash
    g_hash_table_insert(ladspa_descriptors,
                        GINT_TO_POINTER(type),
                        (gpointer)psDescriptor);

    g_free (typeinfo.type_name);
  }
}

GstPlugin*
plugin_init (GModule *module)
{
  ladspa_descriptors = g_hash_table_new(NULL,NULL);

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  ladspa_plugin = gst_plugin_new("gstladspa");
  g_return_val_if_fail(ladspa_plugin != NULL, NULL);

  LADSPAPluginSearch(ladspa_describe_plugin);

  return ladspa_plugin;
}
