/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

//#define DEBUG_ENABLED
#include <median.h>


static GstElementDetails median_details = {
  "Median effect",
  "Filter/Effect",
  "apply a median filter to an image",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};


/* Median signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_ACTIVE,
  ARG_FILTERSIZE
};

static void	gst_median_class_init	(GstMedianClass *klass);
static void	gst_median_init		(GstMedian *median);

static void	median_9		(unsigned char *src, unsigned char *dest, int height, int width);
static void	gst_median_chain	(GstPad *pad, GstBuffer *buf);

static void	gst_median_set_arg	(GtkObject *object, GtkArg *arg, guint id);
static void	gst_median_get_arg	(GtkObject *object, GtkArg *arg, guint id);

static GstElementClass *parent_class = NULL;
//static guint gst_median_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_median_get_type (void)
{
  static GtkType median_type = 0;

  if (!median_type) {
    static const GtkTypeInfo median_info = {
      "GstMedian",
      sizeof(GstMedian),
      sizeof(GstMedianClass),
      (GtkClassInitFunc)gst_median_class_init,
      (GtkObjectInitFunc)gst_median_init,
      (GtkArgSetFunc)gst_median_set_arg,
      (GtkArgGetFunc)gst_median_get_arg,
      (GtkClassInitFunc)NULL,
    };
    median_type = gtk_type_unique(GST_TYPE_ELEMENT,&median_info);
  }
  return median_type;
}

static void
gst_median_class_init (GstMedianClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstMedian::active", GTK_TYPE_BOOL,
	                    GTK_ARG_READWRITE, ARG_ACTIVE);
  gtk_object_add_arg_type("GstMedian::filtersize", GTK_TYPE_INT,
	                    GTK_ARG_READWRITE, ARG_FILTERSIZE);

  gtkobject_class->set_arg = gst_median_set_arg;
  gtkobject_class->get_arg = gst_median_get_arg;
}

static void
gst_median_init (GstMedian *median)
{
  median->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(median),median->sinkpad);
  gst_pad_set_chain_function(median->sinkpad,gst_median_chain);
  median->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(median),median->srcpad);

  median->filtersize = 5;
  median->active = TRUE;
}

#define PIX_SORT(a,b) { if ((a)>(b)) PIX_SWAP((a),(b)); }
#define PIX_SWAP(a,b) { unsigned char temp=(a);(a)=(b);(b)=temp; }

static void
median_5 (unsigned char *src, unsigned char *dest, int width, int height)
{
  int nLastRow;
  int nLastCol;
  unsigned char p[9];
  int i, j, k;

  nLastCol = width - 1;
  nLastRow = height - 1;

  //copy the top and bottom rows into the result array
  for (i=0; i<width; i++) {
    dest[i] = src[i];
    dest[nLastRow * width + i] = src[nLastRow * width + i];
  }
  dest[i] = src[i];

  nLastCol--;
  nLastRow--;

  /* process the interior pixels */
  i = width + 1;
  for (k=0; k < nLastRow; k++) {
    for (j=0; j < nLastCol; j++, i++) {
      p[0] = src[i-width];
      p[1] = src[i-1];
      p[2] = src[i];
      p[3] = src[i+1];
      p[4] = src[i+width];
      PIX_SORT(p[0],p[1]) ; PIX_SORT(p[3],p[4]) ; PIX_SORT(p[0],p[3]) ;
      PIX_SORT(p[1],p[4]) ; PIX_SORT(p[1],p[2]) ; PIX_SORT(p[2],p[3]) ;
      PIX_SORT(p[1],p[2]) ;
      dest[i] = p[2];
    }
    dest[i] = src[i++];
    dest[i] = src[i++];
  }
  dest[i] = src[i++];
}

static void
median_9 (unsigned char *src, unsigned char *dest, int width, int height)
{
  int nLastRow;
  int nLastCol;
  unsigned char p[9];
  int i, j, k;

  nLastCol = width - 1;
  nLastRow = height - 1;

  //copy the top and bottom rows into the result array
  for (i=0; i<width; i++) {
    dest[i] = src[i];
    dest[nLastRow * width + i] = src[nLastRow * width + i];
  }
  dest[i] = src[i];

  nLastCol--;
  nLastRow--;

  /* process the interior pixels */
  i = width + 1;
  for (k=0; k < nLastRow; k++) {
    for (j=0; j < nLastCol; j++, i++) {
      p[0] = src[i-width-1];
      p[1] = src[i-width];
      p[2] = src[i-width+1];
      p[3] = src[i-1];
      p[4] = src[i];
      p[5] = src[i+1];
      p[6] = src[i+width-1];
      p[7] = src[i+width];
      p[8] = src[i+width+1];
      PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ; 
      PIX_SORT(p[0], p[1]) ; PIX_SORT(p[3], p[4]) ; PIX_SORT(p[6], p[7]) ; 
      PIX_SORT(p[1], p[2]) ; PIX_SORT(p[4], p[5]) ; PIX_SORT(p[7], p[8]) ; 
      PIX_SORT(p[0], p[3]) ; PIX_SORT(p[5], p[8]) ; PIX_SORT(p[4], p[7]) ; 
      PIX_SORT(p[3], p[6]) ; PIX_SORT(p[1], p[4]) ; PIX_SORT(p[2], p[5]) ; 
      PIX_SORT(p[4], p[7]) ; PIX_SORT(p[4], p[2]) ; PIX_SORT(p[6], p[4]) ; 
      PIX_SORT(p[4], p[2]) ;
      dest[i] = p[4];
    }
    dest[i] = src[i++];
    dest[i] = src[i++];
  }
  dest[i] = src[i++];
}

static void
gst_median_chain (GstPad *pad, GstBuffer *buf)
{
  GstMedian *median;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;
  GstMeta *meta;
  int lumsize, chromsize;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  median = GST_MEDIAN (GST_OBJECT_PARENT (pad));

  if (!median->active) {
    gst_pad_push(median->srcpad,buf);
    return;
  }

  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"median: have buffer of %d\n", GST_BUFFER_SIZE(buf));

  outbuf = gst_buffer_new();
  GST_BUFFER_DATA(outbuf) = g_malloc(GST_BUFFER_SIZE(buf));
  GST_BUFFER_SIZE(outbuf) = GST_BUFFER_SIZE(buf);

  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (median->format != ((MetaVideoRaw *)meta)->format ||
        median->width  != ((MetaVideoRaw *)meta)->width ||
        median->height != ((MetaVideoRaw *)meta)->height) {

      median->format = ((MetaVideoRaw *)meta)->format;
      median->width  = ((MetaVideoRaw *)meta)->width;
      median->height = ((MetaVideoRaw *)meta)->height;

      gst_buffer_add_meta(outbuf, meta);
    }
  }
  lumsize = median->width*median->height;
  chromsize = lumsize/4;

  if (median->filtersize == 5) {
    median_5(data, GST_BUFFER_DATA(outbuf), median->width, median->height);
    median_5(data+lumsize, GST_BUFFER_DATA(outbuf)+lumsize, median->width/2, median->height/2);
    median_5(data+lumsize+chromsize, GST_BUFFER_DATA(outbuf)+lumsize+chromsize, median->width/2, median->height/2);
  }
  else {
    median_9(data, GST_BUFFER_DATA(outbuf), median->width, median->height);
    median_9(data+lumsize, GST_BUFFER_DATA(outbuf)+lumsize, median->width/2, median->height/2);
    median_9(data+lumsize+chromsize, GST_BUFFER_DATA(outbuf)+lumsize+chromsize, median->width/2, median->height/2);
  }

  gst_buffer_unref(buf);

  gst_pad_push(median->srcpad,outbuf);

}

static void
gst_median_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstMedian *median;
  gint argvalue;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MEDIAN(object));
  median = GST_MEDIAN(object);

  switch(id) {
    case ARG_FILTERSIZE:
      argvalue = GTK_VALUE_INT(*arg);
      if (argvalue != 5 && argvalue != 9) {
	g_print("median: invalid filtersize (%d), must be 5 or 9\n", argvalue);
      }
      else {
	median->filtersize = argvalue;
      }
      break;
    case ARG_ACTIVE:
      median->active = GTK_VALUE_BOOL(*arg);
      break;
    default:
      break;
  }
}

static void
gst_median_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstMedian *median;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_MEDIAN(object));
  median = GST_MEDIAN(object);

  switch (id) {
    case ARG_FILTERSIZE:
      GTK_VALUE_INT(*arg) = median->filtersize;
      break;
    case ARG_ACTIVE:
      GTK_VALUE_BOOL(*arg) = median->active;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}


GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("median");
  g_return_val_if_fail(plugin != NULL, NULL);

  factory = gst_elementfactory_new("median",GST_TYPE_MEDIAN,
                                   &median_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
