/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "icecastsend.h"

/* elementfactory information */
static GstElementDetails icecastsend_details = {
  "An Icecast  plugin",
  "IcecastSend",
  "Sends data to an icecast server",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

/* IcecastSend signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_IP,
  ARG_PORT,
  ARG_PASSWORD,
};

static GstPadFactory sink_factory = {
  "sink",
  GST_PAD_FACTORY_SINK,
  GST_PAD_FACTORY_ALWAYS,
  GST_PAD_FACTORY_CAPS(
  "icecastsend_sink",
     "audio/mp3"
  ),
  NULL
};

static void			gst_icecastsend_class_init	(GstIcecastSendClass *klass);
static void			gst_icecastsend_init		(GstIcecastSend *icecastsend);

static void			gst_icecastsend_chain		(GstPad *pad, GstBuffer *buf);

static void			gst_icecastsend_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void			gst_icecastsend_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static GstElementStateReturn	gst_icecastsend_change_state	(GstElement *element);

GstPadTemplate *src_template, *sink_template;

static GstElementClass *parent_class = NULL;
//static guint gst_icecastsend_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_icecastsend_get_type(void)
{
  static GtkType icecastsend_type = 0;

  if (!icecastsend_type) {
    static const GtkTypeInfo icecastsend_info = {
      "GstIcecastSend",
      sizeof(GstIcecastSend),
      sizeof(GstIcecastSendClass),
      (GtkClassInitFunc)gst_icecastsend_class_init,
      (GtkObjectInitFunc)gst_icecastsend_init,
      (GtkArgSetFunc)gst_icecastsend_set_arg,
      (GtkArgGetFunc)gst_icecastsend_get_arg,
      (GtkClassInitFunc)NULL,
    };
    icecastsend_type = gtk_type_unique(GST_TYPE_ELEMENT,&icecastsend_info);
  }
  return icecastsend_type;
}

static void
gst_icecastsend_class_init (GstIcecastSendClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstIcecastSend::ip", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_IP);
  gtk_object_add_arg_type("GstIcecastSend::port", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_PORT);
  gtk_object_add_arg_type("GstIcecastSend::password", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_PASSWORD);

  gtkobject_class->set_arg = gst_icecastsend_set_arg;
  gtkobject_class->get_arg = gst_icecastsend_get_arg;

  gstelement_class->change_state = gst_icecastsend_change_state;
}

static void
gst_icecastsend_init (GstIcecastSend *icecastsend)
{
  icecastsend->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(icecastsend),icecastsend->sinkpad);
  gst_pad_set_chain_function(icecastsend->sinkpad,gst_icecastsend_chain);

  icecastsend->ip = g_strdup ("127.0.0.1");
  icecastsend->port = 8000;
  icecastsend->password = g_strdup ("");
}

static void
gst_icecastsend_chain (GstPad *pad, GstBuffer *buf)
{
  GstIcecastSend *icecastsend;
  glong ret;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  icecastsend = GST_ICECASTSEND (GST_OBJECT_PARENT (pad));

  g_return_if_fail (icecastsend != NULL);
  g_return_if_fail (GST_IS_ICECASTSEND (icecastsend));

  ret = shout_send_data (&icecastsend->conn, GST_BUFFER_DATA (buf),
		                             GST_BUFFER_SIZE (buf));
  if (!ret) {
    g_warning ("send error: %i...\n", icecastsend->conn.error);
  }

  shout_sleep (&icecastsend->conn);

  gst_buffer_unref (buf);
}

static void
gst_icecastsend_set_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstIcecastSend *icecastsend;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ICECASTSEND(object));
  icecastsend = GST_ICECASTSEND(object);

  switch(id) {
    case ARG_IP:
      if (icecastsend->ip)
	g_free (icecastsend->ip);
      icecastsend->ip = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    case ARG_PORT:
      icecastsend->port = GTK_VALUE_INT (*arg);
      break;
    case ARG_PASSWORD:
      if (icecastsend->password)
	g_free (icecastsend->password);
      icecastsend->password = g_strdup (GTK_VALUE_STRING (*arg));
      break;
    default:
      break;
  }
}

static void
gst_icecastsend_get_arg (GtkObject *object,GtkArg *arg,guint id)
{
  GstIcecastSend *icecastsend;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_ICECASTSEND(object));
  icecastsend = GST_ICECASTSEND(object);

  switch (id) {
    case ARG_IP:
      GTK_VALUE_STRING(*arg) = icecastsend->ip;
      break;
    case ARG_PORT:
      GTK_VALUE_INT(*arg) = icecastsend->port;
      break;
    case ARG_PASSWORD:
      GTK_VALUE_STRING(*arg) = icecastsend->password;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

static GstElementStateReturn
gst_icecastsend_change_state (GstElement *element)
{
  GstIcecastSend *icecastsend;

  g_return_val_if_fail (GST_IS_ICECASTSEND (element), GST_STATE_FAILURE);

  icecastsend = GST_ICECASTSEND(element);

  GST_DEBUG (0,"state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  switch (GST_STATE_TRANSITION (element)) {
   case GST_STATE_NULL_TO_READY:
     shout_init_connection (&icecastsend->conn);

     icecastsend->conn.ip = icecastsend->ip;
     icecastsend->conn.port = icecastsend->port;
     icecastsend->conn.password = icecastsend->password;

     if (shout_connect (&icecastsend->conn)) {
       g_print ("connected to server...\n");
     }
     else {
       g_warning ("couldn't connect to server... (%i)\n", icecastsend->conn.error);
       shout_disconnect (&icecastsend->conn);
       return GST_STATE_FAILURE;
     }
     break;
   case GST_STATE_READY_TO_NULL:
     shout_disconnect (&icecastsend->conn);
     break;
   default:
     break;
  }

  /* if we haven't failed already, give the parent class a chance to ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("icecastsend");
  g_return_val_if_fail(plugin != NULL, NULL);

  factory = gst_elementfactory_new("icecastsend", GST_TYPE_ICECASTSEND, &icecastsend_details);
  g_return_val_if_fail(factory != NULL, NULL);

  sink_template = gst_padtemplate_new (&sink_factory);
  gst_elementfactory_add_padtemplate (factory, sink_template);

  gst_plugin_add_factory (plugin, factory);

  return plugin;
}
