/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "gstjpegdec.h"
#include "gstjpegenc.h"

/* elementfactory information */
extern GstElementDetails gst_jpegdec_details;
extern GstElementDetails gst_jpegenc_details;

GstPadTemplate *dec_src_template, *dec_sink_template; 
GstPadTemplate *enc_src_template, *enc_sink_template;

static GstCapsFactory jpeg_caps_factory = {
  "jpeg_jpeg",
  "video/jpeg",
  NULL
};

static GstCapsFactory raw_caps_factory = {
  "jpeg_raw",
  "video/raw",
  "fourcc",    GST_PROPS_LIST (
                 GST_PROPS_FOURCC ('Y','V','1','2'),
                 GST_PROPS_FOURCC ('I','Y','U','V'),
                 GST_PROPS_FOURCC ('I','4','2','0')
               ),
  "width",     GST_PROPS_INT_RANGE (16, 4096),
  "height",    GST_PROPS_INT_RANGE (16, 4096),
  NULL
};

GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *dec, *enc;
  GstCaps *raw_caps, *jpeg_caps;

  plugin = gst_plugin_new("jpeg");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the jpegdec element */
  enc = gst_elementfactory_new("jpegenc",GST_TYPE_JPEGENC,
                                   &gst_jpegenc_details);
  g_return_val_if_fail(enc != NULL, NULL);
  gst_plugin_add_factory(plugin,enc);

  raw_caps = gst_caps_register (&raw_caps_factory);
  jpeg_caps = gst_caps_register (&jpeg_caps_factory);

  /* register sink pads */
  enc_sink_template = gst_padtemplate_create ("sink", GST_PAD_SINK, 
		                              GST_PAD_ALWAYS, 
					      g_list_prepend (NULL, raw_caps));
  gst_elementfactory_add_padtemplate (enc, enc_sink_template);

  /* register src pads */
  enc_src_template = gst_padtemplate_create ("src", GST_PAD_SRC, 
		                             GST_PAD_ALWAYS, 
					     g_list_prepend (NULL, jpeg_caps));
  gst_elementfactory_add_padtemplate (enc, enc_src_template);

  gst_plugin_add_factory (plugin, enc);

  /* create an elementfactory for the jpegdec element */
  dec = gst_elementfactory_new("jpegdec",GST_TYPE_JPEGDEC,
                                   &gst_jpegdec_details);
  g_return_val_if_fail(dec != NULL, NULL);
 
  /* register sink pads */
  dec_sink_template = gst_padtemplate_create ("sink", GST_PAD_SINK, 
		                              GST_PAD_ALWAYS, 
					      g_list_prepend (NULL, jpeg_caps));
  gst_elementfactory_add_padtemplate (dec, dec_sink_template);

  /* register src pads */
  dec_src_template = gst_padtemplate_create ("src", GST_PAD_SRC, 
		                             GST_PAD_ALWAYS, 
					     g_list_prepend (NULL, raw_caps));
  gst_elementfactory_add_padtemplate (dec, dec_src_template);
  
  gst_plugin_add_factory (plugin, dec);

  return plugin;
}
