/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

//#define DEBUG_ENABLED
#include "gstmpegaudio.h"

/* elementfactory information */
static GstElementDetails gst_mpegaudio_details = {
  "mpegaudio mp3 encoder",
  "Filter/Encoder/Audio",
  "Uses modified mpegaudio code to encode mp3 streams",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* MpegAudio signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void	gst_mpegaudio_class_init	(GstMpegAudioClass *klass);
static void	gst_mpegaudio_init		(GstMpegAudio *mpegaudio);

static void	gst_mpegaudio_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;
//static guint gst_mpegaudio_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_mpegaudio_get_type (void)
{
  static GtkType mpegaudio_type = 0;

  if (!mpegaudio_type) {
    static const GtkTypeInfo mpegaudio_info = {
      "GstMpegAudio",
      sizeof(GstMpegAudio),
      sizeof(GstMpegAudioClass),
      (GtkClassInitFunc)gst_mpegaudio_class_init,
      (GtkObjectInitFunc)gst_mpegaudio_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    mpegaudio_type = gtk_type_unique(GST_TYPE_ELEMENT,&mpegaudio_info);
  }
  return mpegaudio_type;
}

static void
gst_mpegaudio_class_init (GstMpegAudioClass *klass)
{
  GstElementClass *gstelement_class;

  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);
}

static void
gst_mpegaudio_init (GstMpegAudio *mpegaudio)
{
  /* create the sink and src pads */
  mpegaudio->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(mpegaudio),mpegaudio->sinkpad);
  gst_pad_set_chain_function(mpegaudio->sinkpad,gst_mpegaudio_chain);
  mpegaudio->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(mpegaudio),mpegaudio->srcpad);

  /* initialize the mpegaudio encoder state */
  mpegaudio->encoder = mpegaudio_init_encoder();
  mpegaudio->partialbuf = NULL;
  mpegaudio->partialsize = 0;
}

static void
gst_mpegaudio_chain (GstPad *pad, GstBuffer *buf)
{
  GstMpegAudio *mpegaudio;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;
  guint handled, tohandle;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
//  g_return_if_fail(GST_IS_BUFFER(buf));

  mpegaudio = GST_MPEGAUDIO (GST_OBJECT_PARENT (pad));

  data = (guchar *)GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"gst_mpegaudio_chain: got buffer of %ld bytes in '%s'\n",size,
          GST_OBJECT_NAME (mpegaudio));

  handled = 0;
  tohandle = get_number_of_input_bytes(mpegaudio->encoder);

  if (mpegaudio->partialbuf) {
     mpegaudio->partialbuf = g_realloc(mpegaudio->partialbuf, mpegaudio->partialsize+size);
     memcpy(mpegaudio->partialbuf+mpegaudio->partialsize, data,size);

     data = mpegaudio->partialbuf;
     size += mpegaudio->partialsize;
  }

  GST_DEBUG (0,"need to handle %d bytes\n", tohandle);
  while (handled+tohandle < size) {

    outbuf = gst_buffer_new();
    GST_BUFFER_DATA(outbuf) = g_malloc(tohandle);

    GST_DEBUG (0,"about to encode a frame\n");
    mpegaudio_encode_frame(mpegaudio->encoder,data, GST_BUFFER_DATA(outbuf), (gulong *)&GST_BUFFER_SIZE(outbuf));

    GST_DEBUG (0,"mpegaudio: pushing %d bytes\n", GST_BUFFER_SIZE(outbuf));
    gst_pad_push(mpegaudio->srcpad,outbuf);
    GST_DEBUG (0,"mpegaudio: pushed buffer\n");

    data += tohandle;
    handled += tohandle;
  }

  if (size > handled) {
    GST_DEBUG (0,"mpegaudio: leftover buffer %ld bytes\n", size - handled);

    if (!mpegaudio->partialbuf) {
      mpegaudio->partialbuf = g_malloc(size-handled);
    }
    memcpy(mpegaudio->partialbuf, data, size-handled);
    mpegaudio->partialsize = size-handled;
  }
  else {
    if (mpegaudio->partialbuf) {
      g_free(mpegaudio->partialbuf);
      mpegaudio->partialbuf = NULL;
    }
  }

  gst_buffer_unref(buf);
}


GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the putbits package */
  if (!gst_library_load("gstputbits")) {
    gst_info("mpegaudio:: could not load support library: 'gstputbits'\n");
    return NULL;
  }

  plugin = gst_plugin_new("mpegaudio");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the mpegaudio element */
  factory = gst_elementfactory_new("mpegaudio",GST_TYPE_MPEGAUDIO,
                                   &gst_mpegaudio_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
