/*
  definitions for video library
  Copyright (C) 2000  Martin Vogt

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Library General Public License as published by
  the Free Software Foundation.

  For more information look at the file COPYRIGHT in this package

 */



#ifndef __OUTPUT_VIDEO_LIB_H__
#define __OUTPUT_VIDEO_LIB_H__

#ifndef __OMS__


#define IMG_ALLOC(h,w,f)                                                \
        img_buf_t *img;                                                 \
                                                                        \
        if (!(img = malloc (sizeof (img_buf_t))))                       \
                return NULL;                                            \
                                                                        \
        img->width    = width;                                          \
        img->height   = height;                                         \
        img->format   = format;                                         \
                                                                        \
        /* we only know how to do 4:2:0 planar yuv right now. */        \
        if (!(img->base = malloc (width * height * 3 / 2))) {           \
                free(img);                                              \
                return NULL;                                            \
        }                                                               \
                                                                        \
        return img


#define IMG_FREE(img)                                                   \
        free (img->base);                                               \
        free (img)




//
// We use our own structures
//
typedef struct vo_info_s
{
        /* driver name ("Matrox Millennium G200/G400" */
        const char *name;
        /* short name (for config strings) ("mga") */
        const char *short_name;
        /* author ("Aaron Holtzman <aholtzma@ess.engr.uvic.ca>") */
        const char *author;
        /* any additional comments */
        const char *comment;
} vo_info_t;

typedef struct overlay_buf_struct {
        unsigned char *data;    // 7-4: mixer key, 3-0: color index
        unsigned int x;         // x start of subpicture area
        unsigned int y;         // y start of subpicture area
        int width;              // width of subpicture area
        int height;             // height of subpicture area

        unsigned char clut[4];  // color lookup table
        unsigned char trans[4]; // mixer key table

        unsigned int time_execute;     // time in ms
} overlay_buf_t;




typedef struct img_buf_s {
#if 1
// old vo_image_buffer_t stuff ...
        int width;
        int height;
        uint32_t format;
        uint8_t *base;
        void *private;
#else
        union {                 // YUV or RGB buffer pointers
                struct {
                        uint8_t *y;
                        uint8_t *cr;
                        uint8_t *cb; 
                } yuv;
                struct {
                        uint8_t *r;
                        uint8_t *g;
                        uint8_t *b; 
                } rgb;
                uint8_t *data[3];
        } buf;

        int width;
        int height;
        uint32_t format;

        uint8_t stride;         // Y stride, multiple of 16
                                // U and V are using stride/2
// <unused>
        uint8_t format;         // image format
        uint8_t field_id;       // for interlaced output
        uint32_t PTS;
        // whatever comes here
// </unused>

        uint32_t *mem;          // the allocated memory comes here
        void *private;          // user private data
#endif
} img_buf_t;



typedef struct vo_output_video_attr_struct {
	int width;
	int height;
	uint32_t fullscreen;
	char *title;
} vo_output_video_attr_t;

// I really don't like the plugin_ name
typedef vo_output_video_attr_t plugin_output_video_attr_t;

typedef struct vo_output_video_struct {
        void* priv;
        int (*open)	        (void *plugin, void *name);
        int (*close)            (void *plugin);

	int (*setup)		(vo_output_video_attr_t *attr);

	int  (*draw_frame)	(uint8_t *src[]);
	int  (*draw_slice)	(uint8_t *src[], uint32_t slice_num);
	void (*flip_page)	(void);
	img_buf_t*	(*allocate_image_buffer)(int width, int height,
							 uint32_t format);
	void (*free_image_buffer) (img_buf_t* image);
        // 
        // NOT part in OMS
        //
        const vo_info_t* vo_info;

        int (*overlay)          (overlay_buf_t *overlay_buf, uint16_t id);

} vo_output_video_t;

// NULL terminated array of all drivers 
extern vo_output_video_t* video_out_drivers[];





//
// This macro creates a static variable and initialize it
// with given functions
//


#define LIBVIDEO_EXTERN(x,name) \
static vo_info_t x ## _vo_info = {\
        "",\
        name,\
        "",\
        ""\
};\
vo_output_video_t video_out ##x =\
{       priv:                   &( x ## _priv ),\
        open:                   x ## _open,\
        close:                  x ## _close,\
	setup:                  x ## _setup,\
	draw_frame:             x ## _draw_frame,\
	draw_slice:             x ## _draw_slice,\
	flip_page:              x ## _flip_page,\
	allocate_image_buffer:  x ## _allocate_image_buffer,\
	free_image_buffer:      x ## _free_image_buffer,\
        vo_info:                &(x ## _vo_info),\
        overlay:                x ## _overlay\
};


#endif
// != __OMS__

#ifdef __OMS__

//
// Now the OMS part, we have here another type, although
// it has the same names
// Then we define the OMS Plugin functions as well
//
// Future enhancement: remove the functions in plugin_output_video_t
//                     in OMS and replace it with the LIBVIDEO structure
//

#include <oms/oms.h>
#include <oms/plugin/output_video.h>


// I really don't like the plugin_ name in the video lib
// (Its too generic) We should remove the typdefs in
// output_video in OMS anyway and use them in the video lib.
typedef plugin_output_video_attr_t vo_output_video_attr_t;

// This is needed for the "query" feature in the standalone mpeg2dec
// If we "unify" plugin_output_video_t and vo_output_video_t
// we can get rid of that.
typedef plugin_output_video_t vo_output_video_t;

#define LIBVIDEO_EXTERN(x,name) plugin_output_video_t video_out ## x =\
{       priv:                   & x ## _priv,\
        open:                   x ## _open,\
        close:                  x ## _close,\
	setup:                  x ## _setup,\
	draw_frame:             x ## _draw_frame,\
	draw_slice:             x ## _draw_slice,\
	flip_page:              x ## _flip_page,\
        overlay:                x ## _overlay,\
	allocate_image_buffer:  x ## _allocate_image_buffer,\
	free_image_buffer:      x ## _free_image_buffer\
}; \
int PLUGIN_INIT(video_out ## x ##)  (char *whoami)  \
{                                          \
	pluginRegister (whoami,            \
		PLUGIN_ID_OUTPUT_VIDEO,    \
		name,                      \
		NULL,                      \
		NULL,                      \
		&video_out ## x ##);       \
        LOG( LOG_DEBUG, name "video out: Plugin INIT called."); \
	return 0;                          \
}                                          \
void PLUGIN_EXIT(video_out ## x ##) (void) \
{                                          \
        LOG( LOG_DEBUG, name "video out: Plugin INIT called."); \
}                                          \

#endif
// __OMS__


//
// Things which are currently supported by both
//




#endif
