/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


//#define DEBUG_ENABLED
#include <videoscale.h>



/* elementfactory information */
static GstElementDetails videoscale_details = {
  "Video scaler",
  "Filter/Video/Scaler",
  "Resizes video",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

/* Videoscale signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  /* FILL ME */
};

static void	gst_videoscale_class_init	(VideoscaleClass *klass);
static void	gst_videoscale_init		(Videoscale *videoscale);

static void	gst_videoscale_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_videoscale_get_arg		(GtkObject *object, GtkArg *arg, guint id);

static void	gst_videoscale_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;
//static guint gst_videoscale_signals[LAST_SIGNAL] = { 0 };

GtkType
videoscale_get_type (void)
{
  static GtkType videoscale_type = 0;

  if (!videoscale_type) {
    static const GtkTypeInfo videoscale_info = {
      "Videoscale",
      sizeof(Videoscale),
      sizeof(VideoscaleClass),
      (GtkClassInitFunc)gst_videoscale_class_init,
      (GtkObjectInitFunc)gst_videoscale_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    videoscale_type = gtk_type_unique(GST_TYPE_ELEMENT,&videoscale_info);
  }
  return videoscale_type;
}

static void
gst_videoscale_class_init (VideoscaleClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  gtk_object_add_arg_type("Videoscale::width", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_WIDTH);
  gtk_object_add_arg_type("Videoscale::height", GTK_TYPE_INT,
                           GTK_ARG_READWRITE, ARG_HEIGHT);

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_videoscale_set_arg;
  gtkobject_class->get_arg = gst_videoscale_get_arg;

}

static void
gst_videoscale_init (Videoscale *videoscale)
{
  videoscale->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(videoscale),videoscale->sinkpad);
  gst_pad_set_chain_function(videoscale->sinkpad,gst_videoscale_chain);
  videoscale->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(videoscale),videoscale->srcpad);

  videoscale->targetwidth = -1;
  videoscale->targetheight = -1;
  //videoscale->method = GST_VIDEOSCALE_NEAREST;
  //videoscale->method = GST_VIDEOSCALE_BILINEAR;
  videoscale->method = GST_VIDEOSCALE_POINT_SAMPLE;
}


static void
gst_videoscale_chain (GstPad *pad, GstBuffer *buf)
{
  Videoscale *videoscale;
  GstMeta *meta;
  guchar *data;
  gulong size;
  GstBuffer *outbuf;

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  videoscale = GST_VIDEOSCALE (GST_OBJECT_PARENT (pad));
  data = GST_BUFFER_DATA(buf);
  size = GST_BUFFER_SIZE(buf);

  GST_DEBUG (0,"gst_videoscale_chain: got buffer of %ld bytes in '%s'\n",size,
		              GST_OBJECT_NAME (videoscale));

  outbuf = gst_buffer_new();
  GST_BUFFER_DATA(outbuf) = g_malloc(videoscale->targetwidth*videoscale->targetheight*3);
  GST_BUFFER_SIZE(outbuf) = videoscale->targetwidth*videoscale->targetheight*3;

  /* first deal with video metadata */
  meta = gst_buffer_get_first_meta(buf);
  if (meta) {
    if (videoscale->format != ((MetaVideoRaw *)meta)->format ||
        videoscale->width  != ((MetaVideoRaw *)meta)->width ||
        videoscale->height != ((MetaVideoRaw *)meta)->height) {

      videoscale->format = ((MetaVideoRaw *)meta)->format;
      videoscale->width  = ((MetaVideoRaw *)meta)->width;
      videoscale->height = ((MetaVideoRaw *)meta)->height;

      if (videoscale->scaler) gst_videoscale_destroy(videoscale->scaler);
      videoscale->scaler = gst_videoscale_new(videoscale->width, videoscale->height, videoscale->targetwidth, videoscale->targetheight,
		      videoscale->format, videoscale->method);

      ((MetaVideoRaw *)meta)->width = videoscale->targetwidth;
      ((MetaVideoRaw *)meta)->height = videoscale->targetheight;

      gst_buffer_add_meta(outbuf, meta);
    }
  }

  gst_videoscale_scale(videoscale->scaler, data, GST_BUFFER_DATA(outbuf));

  GST_DEBUG (0,"gst_videoscale_chain: pushing buffer of %d bytes in '%s'\n",GST_BUFFER_SIZE(outbuf),
		              GST_OBJECT_NAME (videoscale));

  gst_pad_push(videoscale->srcpad, outbuf);

  gst_buffer_unref(buf);
}

static void
gst_videoscale_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  Videoscale *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VIDEOSCALE(object));
  src = GST_VIDEOSCALE(object);

  switch(id) {
    case ARG_WIDTH:
      src->targetwidth = GTK_VALUE_INT(*arg);
      break;
    case ARG_HEIGHT:
      src->targetheight = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void
gst_videoscale_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  Videoscale *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_VIDEOSCALE(object));
  src = GST_VIDEOSCALE(object);

  switch(id) {
    case ARG_WIDTH:
      GTK_VALUE_INT(*arg) = src->targetwidth;
      break;
    case ARG_HEIGHT:
      GTK_VALUE_INT(*arg) = src->targetheight;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}


GstPlugin*
plugin_init (GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  /* this filter needs the riff parser */
  if (!gst_library_load("gstvideoscale")) {
    gst_info("mpeg2play:: could not load support library: 'gstvideoscale'\n");
    return NULL;
  }

  plugin = gst_plugin_new("videoscale");
  g_return_val_if_fail(plugin != NULL, NULL);

  /* create an elementfactory for the videoscale element */
  factory = gst_elementfactory_new("videoscale",GST_TYPE_VIDEOSCALE,
                                   &videoscale_details);
  g_return_val_if_fail(factory != NULL, NULL);
  gst_plugin_add_factory(plugin,factory);

  return plugin;
}
