/* $Id: gtkoverlayimage.h,v 1.2 2000/04/09 21:36:56 wtay Exp $
 *
 * GtkOverlayImage
 * Copyright (C) 1999-2000 Zoltn Sinkovics and Szabolcs Self
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef GTK_OVERLAY_IMAGE_H__
#define GTK_OVERLAY_IMAGE_H__

#include <gdk/gdk.h>
#include <gdk/gdkprivate.h>
#include <gtk/gtkwidget.h>
#include <gst/meta/videoraw.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GTK_OVERLAY_IMAGE(obj)              (GTK_CHECK_CAST ((obj), gtk_overlay_image_get_type (), GtkOverlayImage))
#define GTK_OVERLAY_IMAGE_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), gtk_overlay_image_get_type (), GtkOverlayImageClass))
#define GTK_IS_OVERLAY_IMAGE(obj)          (GTK_CHECK_TYPE (obj, gtk_overlay_image_get_type ()))
#define GTK_IS_OVERLAY_IMAGE_CLASS(klass)        (GTK_CHECK_CLASS_TYPE ((klass), gtk_overlay_image_get_type ()))
  
#define GTK_OVERLAY_IMAGE_MEM(obj)      (GTK_OVERLAY_IMAGE(obj)->image->mem)
#define GTK_OVERLAY_IMAGE_WIDTH(obj)    (GTK_OVERLAY_IMAGE(obj)->image->width)
#define GTK_OVERLAY_IMAGE_HEIGHT(obj)    (GTK_OVERLAY_IMAGE(obj)->image->height)
#define GTK_OVERLAY_IMAGE_PIXEL_SIZE(obj) (((GdkImagePrivate *)GTK_OVERLAY_IMAGE(obj)->image)->ximage->bits_per_pixel)
#define GTK_OVERLAY_IMAGE_DEPTH(obj) (((GdkImagePrivate *)GTK_OVERLAY_IMAGE(obj)->image)->ximage->depth)
#define GTK_OVERLAY_IMAGE_BYTE_ORDER(obj) (GTK_OVERLAY_IMAGE(obj)->image->byte_order)
#define GTK_OVERLAY_IMAGE_TYPE(obj) (GTK_OVERLAY_IMAGE(obj)->image->type)

typedef struct _GtkOverlayImage      GtkOverlayImage;
typedef struct _GtkOverlayImageClass GtkOverlayImageClass;

struct _GtkOverlayImage
{
  GtkWidget widget;

  GdkImage *image;

  guint refresh_handler;
  gboolean refreshed; 
  gboolean moved;
  gboolean want_config;
  gboolean garbage_collect;
  gboolean drawing_enabled;
  gint newx, newy, newwidth, newheight;
  GdkVisibilityState visibility;
  MetaOverlay meta_overlay;
};

struct _GtkOverlayImageClass
{
  GtkWidgetClass parent_class;

  void (*overlay_change) (GtkWidget *widget, MetaOverlay *raw);
};

GtkType    gtk_overlay_image_get_type     (void);
GtkWidget* gtk_overlay_image_new          (void);
void       gtk_overlay_image_resize       (GtkOverlayImage *widget, gint width, gint height);

#ifdef __cplusplus
}
#endif /* __cplusplus */



#endif /* GTK_OVERLAY_IMAGE_H__ */

