/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "config.h"

#include "synaesthesia.h"
#include "core.h"

static gboolean gst_synaesthesia_start(GstElement *element);

GstElementDetails gst_synaesthesia_details = {
  "Synaesthesia display",
  "Sink/Visualization",
  "Cool color display based on stereo info",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};

static GstElementClass *parent_class = NULL;
//static guint gst_synaesthesia_signals[LAST_SIGNAL] = { 0 };

/* Synaesthesia signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT,
  ARG_WIDGET,
};

static GstPadFactory sink_factory = {
  "sink",                                       /* the name of the pads */
  GST_PAD_FACTORY_SINK,                         /* type of the pad */
  GST_PAD_FACTORY_ALWAYS,                       /* ALWAYS/SOMETIMES */
  GST_PAD_FACTORY_CAPS(
    "synaesthesia_sink16",                      /* the name of the caps */
    "audio/raw",                                /* the mime type of the caps */
    /* Properties follow: */
    "format",   GST_PROPS_INT (AFMT_S16_LE),
    "depth",    GST_PROPS_INT (16)
    // These properties commented out so that autoplugging works for now:
    // the autoplugging needs to be fixed (caps negotiation needed)
    //,"rate",     GST_PROPS_INT (44100)
    //,"channels", GST_PROPS_INT (2)
  ),
  NULL
};

static void gst_synaesthesia_class_init(GstSynaesthesiaClass *klass);
static void gst_synaesthesia_init(GstSynaesthesia *synaesthesia);

static void gst_synaesthesia_chain(GstPad *pad,GstBuffer *buf);

static void gst_synaesthesia_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_synaesthesia_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstPadTemplate *sink_template;

GtkType
gst_synaesthesia_get_type(void) {
  static GtkType synaesthesia_type = 0;

  if (!synaesthesia_type) {
    static const GtkTypeInfo synaesthesia_info = {
      "GstSynaesthesia",
      sizeof(GstSynaesthesia),
      sizeof(GstSynaesthesiaClass),
      (GtkClassInitFunc)gst_synaesthesia_class_init,
      (GtkObjectInitFunc)gst_synaesthesia_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    synaesthesia_type = gtk_type_unique(GST_TYPE_ELEMENT,&synaesthesia_info);
  }
  return synaesthesia_type;
}

static void
gst_synaesthesia_class_init(GstSynaesthesiaClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstSynaesthesia::width", GTK_TYPE_INT,
                          GTK_ARG_READABLE, ARG_WIDTH);
  gtk_object_add_arg_type("GstSynaesthesia::height", GTK_TYPE_INT,
                          GTK_ARG_READABLE, ARG_HEIGHT);
  gtk_object_add_arg_type("GstSynaesthesia::widget", GTK_TYPE_WIDGET,
                          GTK_ARG_READABLE, ARG_WIDGET);

  gtkobject_class->set_arg = gst_synaesthesia_set_arg;
  gtkobject_class->get_arg = gst_synaesthesia_get_arg;
}

static void
gst_synaesthesia_init(GstSynaesthesia *synaesthesia)
{
  synaesthesia->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad(GST_ELEMENT(synaesthesia), synaesthesia->sinkpad);
  gst_pad_set_chain_function(synaesthesia->sinkpad, gst_synaesthesia_chain);

  gst_synaesthesia_start(GST_ELEMENT(synaesthesia));
}

static void gst_synaesthesia_chain(GstPad *pad,GstBuffer *buf) {
  GstSynaesthesia *syna;
  gint samplecount;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  syna = GST_SYNAESTHESIA(GST_OBJECT_PARENT (pad));
  g_return_if_fail(syna != NULL);
  g_return_if_fail(GST_IS_SYNAESTHESIA(syna));

  samplecount = GST_BUFFER_SIZE(buf) /
                (2 * sizeof(gint16));

//  GST_DEBUG (0,"fading\n");
//  fade(&syna->sp);
  GST_DEBUG (0,"doing effect\n");
  coreGo(&syna->sp,GST_BUFFER_DATA(buf),samplecount);

//  GST_DEBUG (0,"drawing\n");
/*  GST_DEBUG (0,"gdk_draw_indexed_image(%p,%p,%d,%d,%d,%d,%s,%p,%d,%p);\n",
        syna->image->window,
	syna->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,syna->width,syna->height,
	"GDK_RGB_DITHER_NORMAL",
	syna->sp.output,syna->width,
	&syna->cmap);*/
/*  gdk_draw_indexed_image(syna->image->window,
	syna->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,syna->width,syna->height,
	GDK_RGB_DITHER_NORMAL,
	syna->sp.output,syna->width,
	&syna->cmap);*/
  gdk_draw_gray_image(syna->image->window,
	syna->image->style->fg_gc[GTK_STATE_NORMAL],
	0,0,syna->width,syna->height,
	GDK_RGB_DITHER_NORMAL,
	syna->sp.output,syna->width);

  gst_trace_add_entry(NULL,0,buf,"synaesthesia: calculated syna");

  gst_buffer_unref(buf);
}

static void gst_synaesthesia_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstSynaesthesia *synaesthesia;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_SYNAESTHESIA(object));
  synaesthesia = GST_SYNAESTHESIA(object);

  switch(id) {
    case ARG_WIDTH:
      synaesthesia->width = GTK_VALUE_INT(*arg);
      break;
    case ARG_HEIGHT:
      synaesthesia->height = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void
gst_synaesthesia_get_arg(GtkObject *object,GtkArg *arg,guint id)
{
  GstSynaesthesia *synaesthesia;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_SYNAESTHESIA(object));
  synaesthesia = GST_SYNAESTHESIA(object);

  GST_DEBUG (0,"have synaesthesia get_arg(%d), wanting %d\n",id,ARG_WIDGET);

  switch(id) {
    case ARG_WIDTH: {
      GTK_VALUE_INT(*arg) = synaesthesia->width;
      GST_DEBUG (0,"returning width value %d\n",GTK_VALUE_INT(*arg));
      break;
    }
    case ARG_HEIGHT: {
      GTK_VALUE_INT(*arg) = synaesthesia->height;
      GST_DEBUG (0,"returning height value %d\n",GTK_VALUE_INT(*arg));
      break;
    }
    case ARG_WIDGET: {
      GTK_VALUE_OBJECT(*arg) = GTK_OBJECT(synaesthesia->image);
      GST_DEBUG (0,"returning widget value %p\n",GTK_VALUE_OBJECT(*arg));
      break;
    }
    default: {
      arg->type = GTK_TYPE_INVALID;
      GST_DEBUG (0,"returning invalid type\n");
      break;
    }
  }
}



GstPlugin *
plugin_init(GModule *module)
{
  GstPlugin *plugin;
  GstElementFactory *factory;

  plugin = gst_plugin_new("synaesthesia");
  g_return_val_if_fail (plugin != NULL, NULL);

  factory = gst_elementfactory_new("synaesthesia", GST_TYPE_SYNAESTHESIA,
                                   &gst_synaesthesia_details);
  g_return_val_if_fail(factory != NULL,NULL);

  sink_template = gst_padtemplate_new(&sink_factory);
  gst_elementfactory_add_padtemplate(factory, sink_template);

  gst_plugin_add_factory(plugin, factory);

  return plugin;
}

static gboolean
gst_synaesthesia_start(GstElement *element)
{
  GstSynaesthesia *syna;

  g_return_val_if_fail(GST_IS_SYNAESTHESIA(element), FALSE);
  syna = GST_SYNAESTHESIA(element);

  syna->width = 255;
  syna->height = 255;
  syna->starsize = 2;

  coreInit(&syna->sp, syna->width, syna->height);
  setStarSize(&syna->sp, syna->starsize);

  setupPalette(&syna->sp, syna->cmap.colors);

  gdk_rgb_init();
  syna->image = gtk_drawing_area_new();
  GST_DEBUG (0,"image is %p\n",syna->image);
  gtk_drawing_area_size(GTK_DRAWING_AREA(syna->image),
			syna->width,
                        syna->height);
  gtk_widget_show(syna->image);

  GST_DEBUG (0,"started synaesthesia\n");
  return TRUE;
}

