/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <vorbisenc.h>
#include <vorbisdec.h>
#include <gst/meta/audioraw.h>

extern GstElementDetails vorbisenc_details;
extern GstElementDetails vorbisdec_details;

static GstCaps* 	vorbis_typefind 	(GstBuffer *buf, gpointer private);

GstPadTemplate *dec_src_template, *dec_sink_template; 
GstPadTemplate *enc_src_template, *enc_sink_template;

static GstCapsFactory vorbis_caps_factory = {
  "vorbis_vorbis",
  "audio/x-ogg",
  NULL
};

static GstCapsFactory raw_caps_factory = {
  "vorbis_raw",
  "audio/raw",
  "format",   GST_PROPS_INT (AFMT_S16_LE),
  "depth",    GST_PROPS_INT (16),
  "rate",     GST_PROPS_INT_RANGE (11025, 48000),
  "channels", GST_PROPS_INT (2),
  NULL
};

static GstTypeFactory vorbisfactory = {
  "audio/x-ogg",
  ".ogg",
  vorbis_typefind,
};

static GstCaps* 
vorbis_typefind (GstBuffer *buf, gpointer private) 
{
  gulong head = GULONG_FROM_BE (*((gulong *)GST_BUFFER_DATA (buf)));

  if (head  != 0x4F676753)
    return NULL;

  return gst_caps_new ("vorbis_typefind", "audio/x-ogg");
}


GstPlugin*
plugin_init (GModule *module) 
{
  GstPlugin *plugin;
  GstElementFactory *enc, *dec;
  GstCaps *raw_caps, *vorbis_caps;

  plugin = gst_plugin_new ("vorbis");
  g_return_val_if_fail (plugin != NULL, NULL);
  gst_plugin_set_longname (plugin, "The OGG Vorbis Codec");

  /* create an elementfactory for the vorbisenc element */
  enc = gst_elementfactory_new ("vorbisenc", GST_TYPE_VORBISENC,
                                &vorbisenc_details);
  g_return_val_if_fail (enc != NULL, NULL);

  raw_caps = gst_caps_register (&raw_caps_factory);
  vorbis_caps = gst_caps_register (&vorbis_caps_factory);

  /* register sink pads */
  enc_sink_template = gst_padtemplate_create ("sink", GST_PAD_SINK, 
		                              GST_PAD_ALWAYS, 
					      g_list_prepend (NULL, raw_caps));
  gst_elementfactory_add_padtemplate (enc, enc_sink_template);

  /* register src pads */
  enc_src_template = gst_padtemplate_create ("src", GST_PAD_SRC, 
		                             GST_PAD_ALWAYS, 
					     g_list_prepend (NULL, vorbis_caps));
  gst_elementfactory_add_padtemplate (enc, enc_src_template);

  gst_plugin_add_factory (plugin, enc);

  /* create an elementfactory for the vorbisdec element */
  dec = gst_elementfactory_new("vorbisdec",GST_TYPE_VORBISDEC,
                               &vorbisdec_details);
  g_return_val_if_fail(dec != NULL, NULL);
 
  /* register sink pads */
  dec_sink_template = gst_padtemplate_create ("sink", GST_PAD_SINK, 
		                              GST_PAD_ALWAYS, 
					      g_list_prepend (NULL, vorbis_caps));
  gst_elementfactory_add_padtemplate (dec, dec_sink_template);

  /* register src pads */
  dec_src_template = gst_padtemplate_create ("src", GST_PAD_SRC, 
		                             GST_PAD_ALWAYS, 
					     g_list_prepend (NULL, raw_caps));
  gst_elementfactory_add_padtemplate (dec, dec_src_template);
  
  gst_plugin_add_factory (plugin, dec);

  gst_plugin_add_type (plugin, &vorbisfactory);

  return plugin;
}
